{-# LANGUAGE GADTs #-}
module ComputationM where

import Types
import Sources
import Functions

import Unsafe.Coerce (unsafeCoerce)
import Control.Monad.State
import Control.Monad.Writer



data TypedRun a b where
  F :: IsF f a b => f -> TypedRun a b
  FIO :: IsFIO f a b => f -> TypedRun a b
  TInject :: b -> TypedRun () b
  GetListElem :: TypedRun () b
  InList :: ComputationM b -> TypedRun [a] [b]

instance Show (ComputationM a) where
  show _ = "<computation>"

deriving instance (Show a, Show b) => Show (TypedRun a b)

data Token a where
  Token :: Int -> Token a
  NoToken :: Token ()

deriving instance Show (Token a)

data Dependency a b = Dependency (Token a) (TypedRun a b) (Token b)
  deriving (Show)

data DependencyUntyped where
  DependencyUntyped :: Dependency a b -> String -> DependencyUntyped

instance Show DependencyUntyped where
  show (DependencyUntyped _ s) = s

newtype ComputationM' a = ComputationM { unComputationM :: WriterT [DependencyUntyped] (State Int) a }
  deriving (Functor, Applicative, Monad, MonadState Int, MonadWriter [DependencyUntyped])

type ComputationM a = ComputationM' (Token a)

evalComputationM m = evalState (execWriterT (unComputationM m)) 0

genDependency :: (Show u, Show a) => (Token a -> Dependency u a) -> ComputationM a
genDependency g = do
  top <- get
  let top' = top + 1
      target = Token top'
  put top'
  let result = g target
  tell [DependencyUntyped result (show result)]
  pure target

inject :: Show a => a -> ComputationM a
inject x = genDependency (Dependency NoToken (TInject x))

getListElem :: Show a => ComputationM a
getListElem = genDependency (Dependency NoToken GetListElem)

run :: (Show a, Show b, IsF f a b) => f -> Token a -> ComputationM b
run f input = genDependency (Dependency input (F f))

isImageFilename = run IsImageFilename
convertedImageFilename = run ConvertedImageFilename

runIO :: (Show a, Show b, IsFIO f a b) => f -> Token a -> ComputationM b
runIO f input = genDependency (Dependency input (FIO f))

listDirectory = runIO ListDirectory
openImage = runIO OpenImage
convertImage = runIO ConvertImage
saveFile = runIO SaveFile
runPandoc = runIO RunPandoc

mapListTaken :: (Show a, Show b) => ComputationM b -> Token [a] -> ComputationM [b]
mapListTaken f input = genDependency (Dependency input (InList f))
